/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  16.10.2016  mifi  First Version.
*  17.10.2016  mifi  Rework to support 8MHz or 48MHz for the CPU frequency.
**************************************************************************/
#define __TALCPU_C__

/*
 * Reference is made to the following source:
 *
 * SAM D21E / SAM D21G / SAM D21J
 * DATASHEET COMPLETE
 * Atmel-42181I-SAM L21_Datasheet_Complete-03/2016 
 */

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdlib.h>
#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#if ((CPU_FREQUENCY != 8000000) && (CPU_FREQUENCY != 48000000))
   Error: CPU_FREQUENCY must defined to 8000000 or 48000000 by the preprocessor;
#endif

#define CONF_CLOCK_DFLL_MULTIPLY_FACTOR      (48000000 / 32768)
#define CONF_CLOCK_DFLL_MAX_COARSE_STEP_SIZE (0x1f / 8)
#define CONF_CLOCK_DFLL_MAX_FINE_STEP_SIZE   (0xff / 8)

/*
 * Available start-up times for the XOSC32K.
 *
 * Available external 32KHz oscillator start-up times, as a number of external
 * clock cycles.
 */
enum system_xosc32k_startup
{
   /** Wait zero clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_0 = 0,
   /** Wait 32 clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_32,
   /** Wait 2048 clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_2048,
   /** Wait 4096 clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_4096,
   /** Wait 16384 clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_16384,
   /** Wait 32768 clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_32768,
   /** Wait 65536 clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_65536,
   /** Wait 131072 clock cycles until the clock source is considered stable */
   SYSTEM_XOSC32K_STARTUP_131072
};

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/* Default System Clock Frequency (Core Clock) */
uint32_t SystemCoreClock = 1000000UL;

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static uint32_t dHiResPeriod = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

#if (48000000 == CPU_FREQUENCY)
/*************************************************************************/
/*  WaitGCLKSync                                                         */
/*                                                                       */
/*  Wait for the complete synchronization of registers between the       */
/*  clock domains.                                                       */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WaitGCLKSync (void)
{
   while(GCLK->STATUS.reg & GCLK_STATUS_SYNCBUSY)
   {
      __asm__ volatile ("nop");
   }
} /* WaitGCLKSync */

/*************************************************************************/
/*  WaitClockSrcReady                                                    */
/*                                                                       */
/*  Wait for the clock source is ready.                                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WaitClockSrcReady (uint32_t dSource)
{
   while(0 == (SYSCTRL->PCLKSR.reg & dSource))
   {
      __asm__ volatile ("nop");
   }
} /* WaitClockSrcReady */

/*************************************************************************/
/*  SetSysClock48MHz                                                     */
/*                                                                       */
/*  Set the system clock to 48MHz.                                       */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void SetSysClock48MHz (void)
{
   uint32_t dValue;

   /* Prevent lint warnings */
   dValue = (uint32_t)SYSTEM_XOSC32K_STARTUP_0     |
            (uint32_t)SYSTEM_XOSC32K_STARTUP_32    |
            (uint32_t)SYSTEM_XOSC32K_STARTUP_2048  |
            (uint32_t)SYSTEM_XOSC32K_STARTUP_4096  |
            (uint32_t)SYSTEM_XOSC32K_STARTUP_16384 |
            (uint32_t)SYSTEM_XOSC32K_STARTUP_32768 |
            (uint32_t)SYSTEM_XOSC32K_STARTUP_65536 |
            (uint32_t)SYSTEM_XOSC32K_STARTUP_131072;
   (void)dValue;            
   
   /* 
    * Set NVM Read Wait States.
    * See page 980, chapter 37.10 NVN Characteristics.
    * For 48MHz 1 NVM Wait States are needed.
    */
   NVMCTRL->CTRLB.bit.RWS = 1;
   
   /* 
    * We need the following setup to generate a 48MHz GCLK_MAIN frequency:
    *
    * XOSC32K -> GCLK Generator 1 -> GCLK Multiplexer 0 -> DFLL48M -> GCLK Generator 0
    */

   /*
    * Reset the Generic Clock Controller first.
    *
    * Due to synchronization, there is a delay from writing CTRL.SWRST until 
    * the reset is complete. CTRL.SWRST and STATUS.SYNCBUSY will both be
    * cleared when the reset is complete.
    */
   GCLK->CTRL.reg = GCLK_CTRL_SWRST;
   
   while( (GCLK->CTRL.reg & GCLK_CTRL_SWRST)        && 
          (GCLK->STATUS.reg & GCLK_STATUS_SYNCBUSY) )
   {
      __asm__ volatile ("nop");
   }
   
   /*
    * See page 126, chapter 15.6.2.1.  Initialization.
    * Setup GCLK Generator 1, use XOSC32K for the source.
    * No division of clock source. 
    */
   
   /*******************************/
   /* Enable clock source XOSC32K */
   /*******************************/
    
   /* 
    * Enable the clock source first, XOSC32K.
    * See page 181, chapter 17.6.3.  32kHz External Crystal Oscillator (XOSC32K) Operation.
    *
    * Warning:  Writing to the XOSC32K.ENABLE bit while writing to other
    *           bits may result in unpredictable behavior. 
    */
   dValue = SYSCTRL_XOSC32K_STARTUP(SYSTEM_XOSC32K_STARTUP_16384) |
            SYSCTRL_XOSC32K_EN32K                                 |
            SYSCTRL_XOSC32K_XTALEN;
   SYSCTRL->XOSC32K.reg = (uint16_t)dValue;

   /* Enable the XOSC32K now */   
   SYSCTRL->XOSC32K.reg |= SYSCTRL_XOSC32K_ENABLE;
   WaitClockSrcReady(SYSCTRL_PCLKSR_XOSC32KRDY);

   /******************************************************/
   /* Connect clock source XOSC32K with GCLK Generator 1 */
   /******************************************************/
   
   /* Set Generic clock generator 1 division register */
   GCLK->GENDIV.reg = (1<<GCLK_GENDIV_ID_Pos); 
   WaitGCLKSync();
   
   /* Set Generic clock generator 1 control register */
   GCLK->GENCTRL.reg = GCLK_GENCTRL_GENEN       |
                       GCLK_GENCTRL_SRC_XOSC32K |  /* Soure */
                       (1<<GCLK_GENDIV_ID_Pos);    /* GCLK Generator ID */
   WaitGCLKSync();

   /****************************************************/
   /* Connect GCLK Generator 1 with GCLK Multiplexer 0 */
   /****************************************************/
   
   /* Set Generic Clock Control register for Multiplexer 0 */
   GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN     |
                       GCLK_CLKCTRL_GEN_GCLK1 | /* Source */
                       GCLK_CLKCTRL_ID_DFLL48;  /* Destination */ /*lint !e845*/
   WaitGCLKSync();

   /*******************************/
   /* Enable clock source DFLL48M */
   /*******************************/
   
   /*
    * Errata 9905:
    * The DFLL clock must be requested before being configured otherwise
    * a write access to a DFLL register can freeze the device.
    *
    * Fix/Workaround: Write a zero to the DFLL ONDEMAND bit in the DFLLCTRL
    *                 register before configuring the DFLL module.
    */      
    
   /*
    * Closed-Loop Operation, see page 184, chapter 17.6.7.1. Basic Operation
    */

   /* Disable ONDEMAND mode while writing configurations */                
   SYSCTRL->DFLLCTRL.reg = 0;
   WaitClockSrcReady(SYSCTRL_PCLKSR_DFLLRDY);

   /* Set DFLL48M Multiplier register */
   SYSCTRL->DFLLMUL.reg = SYSCTRL_DFLLMUL_CSTEP( CONF_CLOCK_DFLL_MAX_COARSE_STEP_SIZE ) |
                          SYSCTRL_DFLLMUL_FSTEP( CONF_CLOCK_DFLL_MAX_FINE_STEP_SIZE )   |
                          SYSCTRL_DFLLMUL_MUL( CONF_CLOCK_DFLL_MULTIPLY_FACTOR );
   WaitClockSrcReady(SYSCTRL_PCLKSR_DFLLRDY);

   /* Disable ONDEMAND mode while writing configurations */                
   SYSCTRL->DFLLCTRL.reg = 0;
   WaitClockSrcReady(SYSCTRL_PCLKSR_DFLLRDY);

   /* Set DFLL48M Control register */
   dValue = SYSCTRL_DFLLCTRL_WAITLOCK |
            SYSCTRL_DFLLCTRL_QLDIS    | 
            SYSCTRL_DFLLCTRL_MODE     |
            SYSCTRL_DFLLCTRL_ENABLE;   
   SYSCTRL->DFLLCTRL.reg = (uint16_t)dValue;
   WaitClockSrcReady(SYSCTRL_PCLKSR_DFLLRDY);

   /* Wait for Coarse, Fine and DFLL ready */
   dValue = SYSCTRL_PCLKSR_DFLLLCKC |
            SYSCTRL_PCLKSR_DFLLLCKF | 
            SYSCTRL_PCLKSR_DFLLRDY;
   WaitClockSrcReady(dValue);

   /******************************************************/
   /* Connect clock source DFLL48M with GCLK Generator 0 */
   /******************************************************/
   
   /* Set Generic clock generator 0 division register */
   GCLK->GENDIV.reg = (0<<GCLK_GENDIV_ID_Pos); 
   WaitGCLKSync();

   /* Set Generic clock generator 0 control register */
   GCLK->GENCTRL.reg = //GCLK_GENCTRL_OE          | 
                       GCLK_GENCTRL_IDC         |  
                       GCLK_GENCTRL_GENEN       |
                       GCLK_GENCTRL_SRC_DFLL48M |  /* Soure */
                       (0<<GCLK_GENCTRL_ID_Pos);   /* GCLK Generator ID */ /*lint !e845*/
   WaitGCLKSync();
   
#if 0 /* Check output frequency of Generic clock generator 0, PB22 */
   PORTB->PINCFG[22].reg = 0x01;  /* Enable MUXEN */
   PORTB->PMUX[11].reg   = 0x07;  /* Peripheral function H selected */
#endif   

   SystemCoreClock = CPU_FREQUENCY;
   
} /* SetSysClock48MHz */
#endif /* (48000000 == CPU_FREQUENCY) */

/*************************************************************************/
/*  NewSysTick_Config                                                    */
/*                                                                       */
/*  Based on the "core_cm4.h" version, but an AHB clock divided by 8 is  */
/*  used for the SysTick clock source.                                   */
/*                                                                       */
/*  The function initializes the System Timer and its interrupt, and     */
/*  starts the System Tick Timer. Counter is in free running mode to     */
/*  generate periodic interrupts.                                        */
/*                                                                       */
/*  In    : ticks                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static uint32_t NewSysTick_Config (uint32_t ticks)
{
   /* Reload value impossible */
   if ((ticks - 1) > SysTick_LOAD_RELOAD_Msk)  return (1);

   /* Set reload register */
   SysTick->LOAD = ticks - 1;
  
   /* Set Priority for Systick Interrupt */
   tal_CPUIrqSetPriority(SysTick_IRQn, SYSTICK_PRIO);
  
   /* Load the SysTick Counter Value */
   SysTick->VAL = 0;
   
   /* 
    * SysTick IRQ and SysTick Timer must be
    * enabled with tal_CPUSysTickStart later.
    */
   
   return(ticks);
} /* NewSysTick_Config */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  tal_CPUInit                                                          */
/*                                                                       */
/*  "Initialize" the CPU.                                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUInit (void)
{
#if (48000000 == CPU_FREQUENCY)
   SetSysClock48MHz();
#endif 

   /* Update clock info */
   SystemCoreClockUpdate(); 
   
   /* Init SysTick */   
   dHiResPeriod = NewSysTick_Config(SystemCoreClock / OS_TICKS_PER_SECOND);

} /* tal_CPUInit */

/*************************************************************************/
/*  tal_CPUSysTickStart                                                  */
/*                                                                       */
/*  Start the SysTick.                                                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUSysTickStart (void)
{
   /* Enable SysTick IRQ and SysTick Timer, use System clock */
   SysTick->CTRL = SysTick_CTRL_CLKSOURCE_Msk | 
                   SysTick_CTRL_TICKINT_Msk   | 
                   SysTick_CTRL_ENABLE_Msk;
                   
} /* tal_CPUSysTickStart */

/*************************************************************************/
/*  tal_CPUIrqEnable                                                     */
/*                                                                       */
/*  Enable the given IRQ.                                                */
/*                                                                       */
/*  In    : IRQ                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqEnable (int IRQ)
{
   NVIC_EnableIRQ((IRQn_Type)IRQ);
} /* tal_CPUIrqEnable */

/*************************************************************************/
/*  tal_CPUIrqDisable                                                    */
/*                                                                       */
/*  Disable the given IRQ.                                               */
/*                                                                       */
/*  In    : IRQ                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqDisable (int IRQ)
{
   NVIC_DisableIRQ((IRQn_Type)IRQ);
} /* tal_CPUIrqDisable */

/*************************************************************************/
/*  tal_CPUIrqDisableAll                                                 */
/*                                                                       */
/*  Disable all interrupts.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqDisableAll (void)
{
   __disable_irq();
} /* tal_CPUIrqDisableAll */

/*************************************************************************/
/*  tal_CPUIrqSetPriority                                                */
/*                                                                       */
/*  Set priority of the given IRQ.                                       */
/*                                                                       */
/*  In    : IRQ, Priority                                                */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqSetPriority (int IRQ, int Priority)
{
   NVIC_SetPriority((IRQn_Type)IRQ, (uint32_t)Priority);
} /* tal_CPUIrqSetPriority */

/*************************************************************************/
/*  tal_CPUStatGetHiResPeriod                                            */
/*                                                                       */
/*  Return the HiResPeriod value.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: HiResPeriod                                                  */
/*************************************************************************/
uint32_t tal_CPUStatGetHiResPeriod (void)
{
   return(dHiResPeriod);
} /* tal_CPUStatGetHiResPeriod */

/*************************************************************************/
/*  tal_CPUStatGetHiResCnt                                               */
/*                                                                       */
/*  Return the HiRes counter.                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: HiRes counter                                                */
/*************************************************************************/
uint32_t tal_CPUStatGetHiResCnt (void)
{
   uint32_t dValue;
   
   /* Get milliseconds */
   dValue  = (OSTimeGet() << 16);
   
   /* The SysTick counts down from HiResPeriod, therefore HiResPeriod - X */
   dValue |= (uint16_t)(dHiResPeriod - SysTick->VAL);
   
   return(dValue);
} /* tal_CPUStatGetHiResCnt */

/*************************************************************************/
/*  tal_CPUGetFrequencyCPU                                               */
/*                                                                       */
/*  Return the clock frequency of the CPU in MHz.                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Frequency                                                    */
/*************************************************************************/
uint32_t tal_CPUGetFrequencyCPU (void)
{
   uint32_t dFrequency;
   
   SystemCoreClockUpdate();   
   dFrequency = SystemCoreClock;
   
   return(dFrequency);
} /* tal_CPUGetFrequencyCPU */

/*************************************************************************/
/*  SysTick_Handler                                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void SysTick_Handler (void)
{
   TAL_CPU_IRQ_ENTER();
   
   OSTimerCallback();
   
   TAL_CPU_IRQ_EXIT();
} /* SysTick_Handler */

/*************************************************************************/
/*  SystemCoreClockUpdate                                                */
/*                                                                       */
/*  Updates the SystemCoreClock with current core Clock.                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void SystemCoreClockUpdate (void)
{
   SystemCoreClock = CPU_FREQUENCY;
} /* SystemCoreClockUpdate */

/*** EOF ***/
